<?php

namespace Drupal\event_scheduler_publish;

use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Entity\ContentEntityBase;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\event_scheduler\EventSchedulerDatabaseInterface;
use Drupal\event_scheduler\EventSchedulerDispatcher;
use Drupal\event_scheduler_publish\Event\EntityPublishEvent;
use Drupal\event_scheduler_publish\Event\EntityScheduledPublishEvent;
use Drupal\event_scheduler_publish\Event\EntityScheduledUnpublishEvent;
use Drupal\event_scheduler_publish\Event\EntityUnpublishEvent;

/**
 * Class SchedulerPublishUtils.
 */
class SchedulerPublishUtils implements SchedulerPublishUtilsInterface {

  /**
   * Drupal\event_scheduler\EventSchedulerDispatcher definition.
   *
   * @var \Drupal\event_scheduler\EventSchedulerDispatcher
   */
  protected $dispatcher;

  /**
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * @var \Drupal\event_scheduler\EventSchedulerDatabaseInterface
   */
  protected $database;

    /**
     * @var \DateTimeZone
     */
    protected $tz;

    /**
     * @var int
     */
    protected $tzOffset = null;

    /**
     * Constructs a new SchedulerPublishUtils object.
     *
     * @param \Drupal\Core\Session\AccountInterface $account
     * @param \Drupal\event_scheduler\EventSchedulerDispatcher $dispatcher
     * @param \Drupal\event_scheduler\EventSchedulerDatabaseInterface $database
     * @param \Drupal\Core\Logger\LoggerChannelInterface $logger
     * @param \Drupal\Component\Datetime\TimeInterface $time
     */
  public function __construct(AccountInterface $account,
      EventSchedulerDispatcher $dispatcher,
                              EventSchedulerDatabaseInterface $database,
                              LoggerChannelInterface $logger
    ) {
    $this->dispatcher = $dispatcher;
    $this->database = $database;
    $this->logger = $logger;

    $userTimezone = function_exists('drupal_get_user_timezone')
        ? drupal_get_user_timezone()
        : date_default_timezone_get();

      $this->tz = new \DateTimeZone($userTimezone);
  }

    /**
     * @inheritDoc
     *
     * @throws \Exception
     */
  public function insertEvents(ContentEntityBase $entity) {
    $currentTime = (new DrupalDateTime('now', $this->tz))->getTimestamp();
    $entityTag = "{$entity->getEntityTypeId()}:{$entity->id()}";

    // Remove these first, just in case they won't be recreated.
    $this->removeExistingEvent(EntityScheduledPublishEvent::NAME, $entityTag);
    $this->removeExistingEvent(EntityScheduledUnpublishEvent::NAME, $entityTag);

    if ($entity->hasField('field_scheduled_start')) {
      $startTime = $this->getTime($entity->get('field_scheduled_start')->value);

      if ($startTime < $currentTime) {
        // The start has already passed, send a publish-entity event now.
        $event = new EntityPublishEvent();
      }
      else {
        // The start time hasn't arrived yet, so schedule it ahead.
        /** @var EntityScheduledPublishEvent $event */
        $event = (new EntityScheduledPublishEvent())
          ->initialise($startTime + $this->tzAdjust(), $entityTag);
      }

      $event->setEntityType($entity->getEntityTypeId())
            ->setBundle($entity->bundle())
            ->setEntityId($entity->id());

      $this->dispatcher->dispatch($event::NAME, $event);
    }

    if ($entity->hasField('field_scheduled_end')) {
      $endTime = $this->getTime($entity->get('field_scheduled_end')->value, TRUE);

      if ($endTime < $currentTime) {
        // The end has already passed, send an unpublish-entity event now.
          $this->removeExistingEvent(EntityScheduledPublishEvent::NAME, $entityTag);
        $event = new EntityUnpublishEvent();
      }
      else {
        // The end time hasn't arrived yet, so schedule it ahead.
        /** @var EntityScheduledUnpublishEvent $event */
        $event = (new EntityScheduledUnpublishEvent())
          ->initialise($endTime + $this->tzAdjust(), $entityTag);
      }

      $event->setEntityType($entity->getEntityTypeId())
            ->setBundle($entity->bundle())
            ->setEntityId($entity->id());

      $this->dispatcher->dispatch($event::NAME, $event);
    }
  }

  /**
   * Get a timestamp from what we are given.
   *
   * If it's midnight and the end time, make it the end of the day.
   *
   * @param string $timeValue
   *
   * @param bool $isEnd
   *
   * @return int
   */
  protected function getTime(string $timeValue, $isEnd = FALSE): int {
      // Get the date/time object adjusting for timezone;
      $timestamp = (new DrupalDateTime($timeValue))->getTimestamp();

    // Is this exactly midnight?
    if (($timestamp % 86400) === 0) {
      if ($isEnd) {
        // If it's the end time, set it to one second to midnight.
        $timestamp += 86399;
      }
    }

    return $timestamp;
  }

    /**
     * Get the time offset in seconds.
     *
     * Counter intuitive but we need double the amount to compensate.
     *
     * @return int
     *
     * @throws \Exception
     */
  protected function tzAdjust(): int {
      if ($this->tzOffset === null) {
          $defaultDateTime = new \DateTime();
          $this->tzOffset = $this->tz->getOffset($defaultDateTime);
      }
      return $this->tzOffset * 2;
  }

  /**
   * Remove a matching event from the event scheduler database.
   *
   * @param string $eventName
   * @param string $entityTag
   */
  protected function removeExistingEvent(string $eventName, string $entityTag): void {
    $this->database->delete([
      'name' => ['value' => $eventName],
      'tag' => ['value' => $entityTag],
    ]);
  }

  /**
   * @inheritDoc
   */
  public function updateEvents(ContentEntityBase $entity) {
    /** @var ContentEntityBase $original */
    $original = $entity->original;
    $changed = false;

    if ($entity->hasField('field_scheduled_start')) {
      if ($original->hasField('field_scheduled_start')) {
        $changed = $entity->get('field_scheduled_start')->value !== $original->get('field_scheduled_start')->value;
      }
      else {
        // The start field has just been added - that's a change.
        $changed  = true;
      }
    }
    elseif ($original->hasField('field_scheduled_start')) {
      // The start field has been removed, so that's a change too.
      $changed = TRUE;
    }

    if (!$changed && $entity->hasField('field_scheduled_end')) {
      if ($original->hasField('field_scheduled_end')) {
        $changed = $entity->get('field_scheduled_end')->value !== $original->get('field_scheduled_end')->value;
      }
      else {
        // The end field has just been added - that's a change.
        $changed = TRUE;
      }
    }
    else if ($original->hasField('field_scheduled_end')) {
      // The end field has been removed, so that's a change too.
      $changed = TRUE;
    }

    if ($changed) {
      $this->insertEvents($entity);
    }
  }

}
